# Best State parent class

import math


def init():
    """
    Initialization of variables. This function is used in subclasses.
    """
    solution = [] # The list containing the lists of switched-on machines and their information.
    solutionNumber = [] # The list containing the number of machines to be switched-on.
    possibleToSwitchOn = True
    return possibleToSwitchOn, solution, solutionNumber


class BestStateClass:
    def __init__(self, listMachine, expe, resultHeuristicExists):
        """
        The BestStateClass parent class.

        :param listMachine: List containing the machine objects (number, static power, states and name).
        :param expe: Experience object (powerStart, powerEnd et byStep).
        :param resultHeuristicExists: Boolean. True if the result files of this experiment with these machines
        already exist, False if not.
        """
        self.listMachine = listMachine
        self.expe = expe
        self.resultHeuristicExists = resultHeuristicExists
        self.objective = []
        self.runtime = -1

    @staticmethod
    def chooseMachineAndState(ratioMachine, possibleToSwitchOn):
        """
        Function that determines the machine & DVFS state to switch-on in the ratioMachine list.

        :param ratioMachine: List of machine combinations - DVFS states giving the ratio with static power.
        :param possibleToSwitchOn: Boolean variable indicating if it is possible to switch-on a machine.
        """
        chosenMachine = -1 # Variable that determines the type of machine to be switched-on.
        chosenState = -1 # Variable that determines the DVFS state of machine to be switched-on.
        if not ratioMachine:
            possibleToSwitchOn = False
        else:
            bestRatio = math.inf
            for ratio in ratioMachine:
                if ratio[2] < bestRatio:
                    chosenMachine = ratio[0]
                    chosenState = ratio[1]
                    bestRatio = ratio[2]
        return chosenMachine, chosenState, possibleToSwitchOn

    def computeRatiosBase(self, numberMachineAvailable):
        """
        Function computing for each machine type and for each DVFS state the ratio (with static power) in the case
        where the available power is higher than the maximum power that can be requested by the most consuming machine.

        :param numberMachineAvailable: Number of machines available for each type of machine.
        """
        ratioMachine = []
        for k in range(len(self.listMachine)):
            if numberMachineAvailable[k] > 0:
                for i in range(1, len(self.listMachine[k].states)):
                    dynamicPowerMachine = self.listMachine[k].states[i][0] * self.listMachine[k].states[i][1]
                    computingPowerMachine = self.listMachine[k].states[i][0]
                    ratioMachine.append([k, i, (self.listMachine[k].static + dynamicPowerMachine) /
                                         computingPowerMachine])
        return ratioMachine

    def computeRatios(self, power, numberMachineAvailable):
        """
        Function computing for each type of machine and for each DVFS state the ratio (with static power) in the case
        where the available power is lower than the maximum power that can be requested by the most consuming machine.

        :param power: The available power.
        :param numberMachineAvailable: Number of machines available for each type of machine.
        """
        ratioMachine = []
        for k in range(len(self.listMachine)):
            if numberMachineAvailable[k] > 0:
                if power > self.listMachine[k].static:
                    for i in range(1, len(self.listMachine[k].states)):
                        power_p = power - self.listMachine[k].static
                        dynamicPowerMachine = min(self.listMachine[k].states[i][0] *
                                                  self.listMachine[k].states[i][1], power_p)
                        computingPowerMachine = dynamicPowerMachine / self.listMachine[k].states[i][1]
                        ratioMachine.append([k, i, (self.listMachine[k].static + dynamicPowerMachine) /
                                             computingPowerMachine])
        return ratioMachine

    def switchOnMachine(self, power, chosenMachine, chosenState, numberMachineAvailable, solution, solutionNumber):
        """
        Function that switches on an additional machine and updates the available power and the number of
        machines available.

        :param power: The available power.
        :param chosenMachine: The chosen machine.
        :param chosenState: The chosen DVFS state of the machine.
        :param numberMachineAvailable: Number of machines available for each type of machine.
        :param solution: The list containing the lists of switched on machines and their information.
        :param solutionNumber: The list containing the number of machines switched on.
        """
        for k in range(len(self.listMachine)):
            if chosenMachine == k:
                power -= self.listMachine[k].static
                dynamicPowerMachine = min(self.listMachine[k].states[chosenState][0] *
                                          self.listMachine[k].states[chosenState][1], power)
                if [k, chosenState, dynamicPowerMachine] not in solution:
                    solution.append([k, chosenState, dynamicPowerMachine])
                    solutionNumber.append(1)
                else:
                    ide = solution.index([k, chosenState, dynamicPowerMachine])
                    solutionNumber[ide] += 1
                power -= dynamicPowerMachine
                numberMachineAvailable[k] -= 1
        return power, numberMachineAvailable, solution, solutionNumber

    @staticmethod
    def checkSolution(solution, solutionNumber):
        """
        Function that checks solutionNumber and removes the cases with 0 machines in the solution and solutionNumber.

        :param solution: The list containing the lists of switched on machines and their information.
        :param solutionNumber: The list containing the number of machines switched on.
        """
        for i in range(len(solutionNumber) - 1, -1, -1):
            if solutionNumber[i] == 0:
                del solutionNumber[i]
                del solution[i]
        return solution, solutionNumber

    def possibleStateChange(self, currentPower, solution):
        """
        Function that generates the DVFS state change possibilities for each machine already switched on.

        :param currentPower: The available power
        :param solution: The list containing the lists of switched on machines and their information.
        """
        ratiosTestMachineState = []
        ratiosTest = []
        for machine in solution:
            for k in range(len(self.listMachine)):
                if machine[0] == k:
                    currentState = machine[1]
                    if currentState < len(self.listMachine[k].states) - 1:
                        for state in range(currentState + 1, len(self.listMachine[k].states)):
                            if [k, currentState, state] not in ratiosTestMachineState:
                                currentDynamicPower = machine[2]
                                currentPower += currentDynamicPower
                                newDynamicPower = min(self.listMachine[k].states[state][0] *
                                                      self.listMachine[k].states[state][1], currentPower)
                                currentPower -= currentDynamicPower
                                newComputingPower = newDynamicPower / self.listMachine[k].states[state][1]
                                if newComputingPower > (currentDynamicPower /
                                                        self.listMachine[k].states[currentState][1]):
                                    ratiosTestMachineState.append([k, currentState, state])
                                    ratiosTest.append([newDynamicPower, self.listMachine[k].states[state][1]])
        return currentPower, ratiosTestMachineState, ratiosTest

    @staticmethod
    def ChooseNewState(ratiosTest):
        """
        Function that determines the lowest ratio.

        :param ratiosTest: List containing the dynamic power and the ratio (without static) of the machines.
        """
        indexBestRatio = -1
        bestRatio = math.inf
        for ratio in ratiosTest:
            if ratio[1] < bestRatio:
                bestRatio = ratio[1]
                indexBestRatio = ratiosTest.index(ratio)
        return indexBestRatio

    @staticmethod
    def applyChangeState(currentPower, solution, solutionNumber, ratiosTestMachineState, ratiosTest,
                         indexBestRatio):
        """
        Function applying the DVFS state change to the chosen machine.

        :param currentPower: The available power.
        :param solution: The list containing the lists of switched on machines and their information.
        :param solutionNumber: The list containing the number of machines switched on.
        :param ratiosTestMachineState: List containing the types of machines switched on, their current DVFS state
        and their candidate DVFS states for reallocation.
        :param ratiosTest: List containing the dynamic power and the ratio (without static) of the machines.
        :param indexBestRatio: Selected list index in ratiosTestMachineState and ratiosTest for the DVFS state change.
        """
        for machine in solution:
            if machine[0] == ratiosTestMachineState[indexBestRatio][0]:
                if machine[1] == ratiosTestMachineState[indexBestRatio][1] and \
                        currentPower + machine[2] - ratiosTest[indexBestRatio][0] >= -1e-8:
                    ide = solution.index(machine)
                    solutionNumber[ide] -= 1
                    currentPower += machine[2]
                    currentPower -= ratiosTest[indexBestRatio][0]
                    if [ratiosTestMachineState[indexBestRatio][0], ratiosTestMachineState[indexBestRatio][2],
                        ratiosTest[indexBestRatio][0]] not in solution:
                        solution.append([ratiosTestMachineState[indexBestRatio][0],
                                         ratiosTestMachineState[indexBestRatio][2], ratiosTest[indexBestRatio][0]])
                        solutionNumber.append(1)
                    else:
                        ide = solution.index([ratiosTestMachineState[indexBestRatio][0],
                                              ratiosTestMachineState[indexBestRatio][2],
                                              ratiosTest[indexBestRatio][0]])
                        solutionNumber[ide] += 1
                    break
        return currentPower, solution, solutionNumber

    def powerAllocationWithoutStatic(self, power, stop, solution, solutionNumber):
        """
        Function that reallocates available power to machines already switched on.

        :param power: The available power.
        :param stop: Boolean variable specifying to stop or not the power allocation.
        :param solution: The list containing the lists of switched on machines and their information.
        :param solutionNumber: The list containing the number of machines switched on.
        """
        currentPower = power
        solution, solutionNumber = self.checkSolution(solution, solutionNumber)
        currentPower, ratiosTestMachineState, ratiosTest = self.possibleStateChange(currentPower, solution)
        if ratiosTestMachineState:
            indexBestRatio = self.ChooseNewState(ratiosTest)
            currentPower, solution, solutionNumber = self.applyChangeState(currentPower, solution, solutionNumber,
                                                                           ratiosTestMachineState, ratiosTest,
                                                                           indexBestRatio)
            power = currentPower
        else:
            stop = True
        return solution, solutionNumber, stop, power

    def getMachineMax(self):
        """
        Function giving the number of available machines.
        """
        numberMachineAvailable = [self.listMachine[k].number for k in range(len(self.listMachine))]
        return numberMachineAvailable

    def computeObjectiveSolution(self, solution, solutionNumber):
        """
        Function that computes the feasible computing power based on the switched on machines, by reading the
        solution and solutionNumber list.

        :param solution: The list containing the lists of switched on machines and their information.
        :param solutionNumber: The list containing the number of machines switched on.
        """
        if not solution:
            return 0
        else:
            objectiveSolution = 0
            i = 0
            for machine in solution:
                for k in range(len(self.listMachine)):
                    if machine[0] == k:
                        objectiveSolution += (machine[2] / self.listMachine[k].states[machine[1]][1]) * \
                                             solutionNumber[i]
                i += 1
            return objectiveSolution
