# Milp class

from pulp import *
import time


class Milp:
    def __init__(self, listMachine, expe, resultMilpExists):
        """
        The Milp class.

        :param listMachine: List containing the machine objects.
        (number, static power, states and name).
        :param expe: Experience object (powerStart, powerEnd et byStep).
        :param resultMilpExists: Boolean. True if the result files of this experiment with these machines already
        exist, False if not.
        """
        self.x = None  # A binary variable specifying the state of the machine
        self.machinePower = None  # Variable specifying the power consumption of the machine
        self.operationsProcessedMachine = None  # Variable specifying the number of operations processed by the machine
        self.operationsProcessedStateMachine = None  # Variable specifying the number of operations processed by
        # the machine at a given DVFS state
        self.prob = None
        self.listMachine = listMachine
        self.expe = expe
        self.resultMilpExists = resultMilpExists
        self.objectifMilp = []
        self.powerAndMachinesUsed = []
        self.detailedSolution = []
        self.xRuntime = []
        self.runtime = -1

    def getObjectifMilp(self):
        return self.objectifMilp

    def getXRuntime(self):
        return self.xRuntime

    def getRuntime(self):
        return self.runtime

    def setProb(self, x):
        self.prob = x

    def getDetailedSolution(self):
        return self.detailedSolution

    def variables(self):
        """
        Definition of the problem variables.
        """
        self.x = []
        self.machinePower = []
        self.operationsProcessedMachine = []
        self.operationsProcessedStateMachine = []
        for k in range(len(self.listMachine)):
            self.x.append(LpVariable.dict("x_" + str(k), (list(range(self.listMachine[k].number)),
                                                          list(range(len(self.listMachine[k].states)))),
                                          lowBound=0, upBound=1, cat="Binary"))
            self.machinePower.append(
                LpVariable.dict("machinePower_" + str(k), list(range(self.listMachine[k].number)),
                                lowBound=0, cat="Continuous"))
            self.operationsProcessedMachine.append(
                LpVariable.dict("operationsProcessedMachine_" + str(k), list(range(self.listMachine[k].number)),
                                lowBound=0, cat="Continuous"))
            self.operationsProcessedStateMachine.append(LpVariable.dict(
                "operationsProcessedStateMachine_" + str(k), (list(range(self.listMachine[k].number)),
                                                              list(range(len(self.listMachine[k].states)))), lowBound=0,
                cat="Continuous"))

    def constraints(self, power):
        """
        Definition of the problem constraints.

        :param power : The available power value considered.
        """
        self.prob += (lpSum(self.operationsProcessedMachine[k][i] for k in range(len(self.listMachine))
                            for i in range(self.listMachine[k].number)), "TotalWorkloadProcessed")
        for k in range(len(self.listMachine)):
            for i in range(self.listMachine[k].number):
                self.prob += lpSum(self.x[k][i, j] for j in range(len(self.listMachine[k].states))) == 1
                for j in range(len(self.listMachine[k].states)):
                    self.prob += self.operationsProcessedStateMachine[k][i, j] <= lpSum(
                        self.x[k][i, j] * self.listMachine[k].states[j][0])
                self.prob += self.operationsProcessedStateMachine[k][i, 0] * self.listMachine[k].states[0][1] + lpSum(
                    self.listMachine[k].static * self.x[k][i, j] + self.operationsProcessedStateMachine[k][i, j] *
                    self.listMachine[k].states[j][1]
                    for j in range(1, len(self.listMachine[k].states))) == self.machinePower[k][i]
                self.prob += lpSum(self.operationsProcessedStateMachine[k][i, j]
                                   for j in range(len(self.listMachine[k].states))) == \
                             self.operationsProcessedMachine[k][i]

        self.prob += lpSum(self.machinePower[k][i] for k in range(len(self.listMachine))
                           for i in range(self.listMachine[k].number)) <= power

    def getPowerAndMachinesUsed(self, power):
        """
        Function that returns the switched on machines, their DVFS state and their dynamic power.

        :param power: The available power.
        """
        listForCurrentPower = []
        listNumber = []
        numberMachineOn = 0
        for k in range(len(self.listMachine)):
            for i in range(self.listMachine[k].number):
                for j in range(len(self.listMachine[k].states)):
                    if j >= 1 and self.x[k][i, j].varValue > 0.5:
                        numberMachineOn += 1
                        dynPower = round(
                            self.operationsProcessedStateMachine[k][i, j].varValue * self.listMachine[k].states[j][1],
                            2)
                        if [k, j, dynPower] not in listForCurrentPower:
                            listForCurrentPower.append([k, j, dynPower])
                            listNumber.append(1)
                        else:
                            inde = listForCurrentPower.index([k, j, dynPower])
                            listNumber[inde] += 1
        listGlobal = [power, numberMachineOn, listNumber, listForCurrentPower]
        return listGlobal

    def computeMilp(self):
        """
        Function that runs the MILP.
        """
        start = time.time()
        timeList = []
        for power in range(self.expe.powerStart, self.expe.powerEnd, self.expe.byStep):
            start2 = time.time()
            print(power, '/', self.expe.powerEnd, ' - by step ', self.expe.byStep)
            self.setProb(LpProblem("ProcessWorkload", LpMaximize))
            self.variables()
            self.constraints(power)
            self.prob.solve(GUROBI(msg=False, gapRel=1e-6))
            self.objectifMilp.append(self.prob.objective.value())
            end2 = time.time()
            self.xRuntime.append(end2 - start2)
            self.detailedSolution.append(self.getPowerAndMachinesUsed(power))
            timeList.append(end2 - start2)
            print("Objective =", self.prob.objective.value())
            print("Time =", end2 - start2)
            print("Average time =", sum(timeList) / len(timeList))
            self.setProb(None)
        end = time.time()
        self.runtime = end - start

    def saveData(self):
        """
        Function that saves the data and results of the problem.
        """
        if not self.resultMilpExists:
            Results = dict()
            self.computeMilp()
            Results['objectifMilp'] = self.objectifMilp
            Results['Time Milp'] = self.xRuntime
            Results['Detailed solution'] = self.detailedSolution
            numberMachine = 0
            for k in range(len(self.listMachine)):
                numberMachine += self.listMachine[k].number
            nom_fichier_solution = "Solution_Milp_with_" + str(len(self.listMachine)) + "_types_and_" + \
                                   str(numberMachine) + "_machine_from_" + str(self.expe.powerStart) + "_to_" + \
                                   str(self.expe.powerEnd) + "_W_by_step_" + str(self.expe.byStep)
            with open(nom_fichier_solution, 'w', encoding='utf-8') as f:
                json.dump(Results, f, indent=4)
